#include <stdint.h>
#include <string.h>

unsigned int CRC16_Modbus(unsigned char* buf, int len)
{
    unsigned int crc = 0xFFFF;
    for (int pos = 0; pos < len; pos++)
    {
        crc ^= (unsigned int)buf[pos]; // XOR-байт в наименьший sig. байт crc
        for (int i = 8; i != 0; i--)   // Повторите цикл над каждым битом
        {
            if ((crc & 0x0001) != 0)   // Если установлен LSB
            {
                crc >>= 1; // Сдвиг вправо и XOR 0xA001
                crc ^= 0xA001;
            }
            else // В противном случае LSB не установлен
            {
                crc >>= 1;    // Сдвиг вправо
            }
        }
    }
    crc = ((crc & 0x00ff) << 8) | ((crc & 0xff00) >> 8); // преобразование младших байтов
    return crc;
}

size_t read_input_status(uint8_t address, uint16_t discret_input, uint16_t quantity_reg, uint8_t* buf, size_t buf_len)
{
	int i=0;
	buf[i] = address;
	i++;
	buf[i] = 0x02;
	i++;
	buf[i] = (discret_input >>8)&0xFF;
	i++;
	buf[i] = discret_input&0xFF;
	i++;
    buf[i] = (quantity_reg >> 8) & 0xFF;
    i++;
    buf[i] = quantity_reg & 0xFF;
    i++;
    uint16_t crc = CRC16_Modbus(buf, i);
    buf[i] = crc & 0xFF;
    i++;
    buf[i] = (crc >> 8) & 0xFF;
    i++;
    return i;
}

int read_input_status_parse(uint8_t* buf, size_t buf_len, uint8_t* meaning)
{
    if (buf_len < 8) {
        // Недостаточная длина ответа
        return 0;
    }
    // Проверка контрольной суммы
    uint16_t crc = CRC16_Modbus(buf, buf_len - 2);
    uint16_t received_crc = buf[buf_len - 2] + (buf[buf_len - 1] << 8);

    if (crc != received_crc) {
        // Контрольная сумма не совпадает
        return 0;
    }

    *meaning = buf[buf_len - 5];
    return 1;
}

size_t read_input_register(uint8_t address, uint16_t input_reg, uint8_t* buf, size_t buf_len)
{
    int i = 0;
    buf[i] = address;
    i++;
    buf[i] = 0x04;
    i++;
    buf[i] = (input_reg >> 8) & 0xFF;
    i++;
    buf[i] = input_reg & 0xFF;
    i++;
    buf[i] = 0x00;
    i++;
    buf[i] = 0x01;
    i++;
    uint16_t crc = CRC16_Modbus(buf, i);
    buf[i] = crc & 0xFF;
    i++;
    buf[i] = (crc >> 8) & 0xFF;
    i++;
    return i;
} 

int read_input_register_parse(uint8_t* buf, size_t buf_len , uint16_t* meaning)
{
    if (buf_len < 5) {
        // Недостаточная длина ответа
        return 0;
    }

    // Проверка контрольной суммы
    uint16_t crc = CRC16_Modbus(buf, buf_len - 2);
    uint16_t received_crc = buf[buf_len - 2] + (buf[buf_len - 1] << 8);

    if (crc != received_crc) {
        // Контрольная сумма не совпадает
        return 0;
    }
    
    *meaning = buf[buf_len - 3] + (buf[buf_len - 4] << 8);
    return 1;
}

size_t forse_single_coil(uint8_t address, uint16_t address_input, int on, uint8_t* buf, size_t buf_len)
{
    int i = 0;
    buf[i] = address;
    i++;
    buf[i] = 0x05;
    i++;
    buf[i] = (address_input >> 8)&0xFF;
    i++;
    buf[i] = address_input & 0xFF;
    i++;
    buf[i] = on ? 0xFF : 0x00;
    i++;
    buf[i] = 0x00;
    i++;
    uint16_t crc = CRC16_Modbus(buf, i);
    buf[i] = crc & 0xFF;
    i++;
    buf[i] = (crc >> 8) & 0xFF;
    i++;
    return i;
}

int forse_single_coil_parse(uint8_t* buf, size_t buf_len)
{
    if (buf_len < 8) {
        // Недостаточная длина ответа
        return 0;
    }

    // Проверка контрольной суммы
    uint16_t crc = CRC16_Modbus(buf, buf_len - 2);
    uint16_t received_crc = buf[buf_len - 2] + (buf[buf_len - 1] << 8);

    if (crc != received_crc) {
        // Контрольная сумма не совпадает
        return 0;
    }

    
    return 1;
}
